"""Here we generate mirror representation of py, Rmd and ipynb files
as py or ipynb, and make sure that these representations minimally
change on new releases.
"""

import os
import sys
import pytest
from testfixtures import compare
import jupytext
from .utils import list_all_notebooks, list_r_notebooks, \
    list_julia_notebooks, list_py_notebooks

jupytext.file_format_version.FILE_FORMAT_VERSION = {}


def mirror_file(nb_file):
    dir, file = os.path.split(nb_file)
    if nb_file.endswith('.py'):
        return os.path.join(dir, 'mirror', file.replace('.py', '.ipynb'))
    if nb_file.endswith('.jl'):
        return os.path.join(dir, 'mirror', file.replace('.jl', '.ipynb'))
    if nb_file.endswith('.Rmd'):
        return os.path.join(dir, 'mirror', file.replace('.Rmd', '.ipynb'))
    return os.path.join(dir, 'mirror', file.replace('.ipynb', '.py'))


def create_if_missing(mirror_file, org_file):
    if not os.path.isfile(mirror_file):
        nb = jupytext.readf(org_file)
        jupytext.writef(nb, mirror_file)


def test_create_if_missing(tmpdir):
    py_file = str(tmpdir.join('notebook.py'))
    ipynb_file = str(tmpdir.join('notebook.ipynb'))
    with open(py_file, 'w') as fp:
        fp.write('1+1\n')

    create_if_missing(ipynb_file, py_file)
    assert os.path.isfile(ipynb_file)


@pytest.mark.skipif(sys.version_info < (3, 6),
                    reason="unordered dict result in changes in chunk options")
@pytest.mark.parametrize('py_file',
                         [py_file for py_file in list_all_notebooks('.py')
                          if py_file.find('notebook_sample') > 0])
def test_py_unchanged_py(py_file):
    with open(py_file, encoding='utf-8') as fp:
        py = fp.read()

    ipynb_file = mirror_file(py_file)
    create_if_missing(ipynb_file, py_file)

    py_ref = jupytext.writes(jupytext.readf(ipynb_file), ext='.py')
    compare(py, py_ref)


@pytest.mark.skipif(sys.version_info < (3, 6),
                    reason="unordered dict result in changes in chunk options")
@pytest.mark.parametrize('rmd_file', list_all_notebooks('.Rmd'))
def test_rmd_unchanged(rmd_file):
    with open(rmd_file, encoding='utf-8') as fp:
        rmd = fp.read()

    ipynb_file = mirror_file(rmd_file)
    create_if_missing(ipynb_file, rmd_file)

    rmd_ref = jupytext.writes(jupytext.readf(ipynb_file), ext='.Rmd')
    compare(rmd, rmd_ref)


@pytest.mark.skipif(sys.version_info < (3, 6),
                    reason="unordered dict result in changes in chunk options")
@pytest.mark.parametrize('nb_file', list_py_notebooks('.ipynb'))
def test_py_unchanged_ipynb(nb_file):
    py_file = mirror_file(nb_file)
    create_if_missing(py_file, nb_file)

    with open(py_file, encoding='utf-8') as fp:
        py_ref = fp.read()

    py = jupytext.writes(jupytext.readf(nb_file), ext='.py')
    compare(py, py_ref)


@pytest.mark.skipif(sys.version_info < (3, 6),
                    reason="unordered dict result in changes in chunk options")
@pytest.mark.parametrize('nb_file', list_r_notebooks('.ipynb'))
def test_R_unchanged_ipynb(nb_file):
    r_file = mirror_file(nb_file).replace('.py', '.R')
    create_if_missing(r_file, nb_file)
    with open(r_file, encoding='utf-8') as fp:
        r_ref = fp.read()

    r = jupytext.writes(jupytext.readf(nb_file), ext='.R')
    compare(r, r_ref)


@pytest.mark.skipif(sys.version_info < (3, 6),
                    reason="unordered dict result in changes in chunk options")
@pytest.mark.parametrize('nb_file', list_julia_notebooks('.ipynb'))
def test_julia_unchanged_ipynb(nb_file):
    julia_file = mirror_file(nb_file).replace('.py', '.jl')
    create_if_missing(julia_file, nb_file)
    with open(julia_file, encoding='utf-8') as fp:
        julia_ref = fp.read()

    julia = jupytext.writes(jupytext.readf(nb_file), ext='.jl')
    compare(julia, julia_ref)


@pytest.mark.skipif(sys.version_info < (3, 6),
                    reason="unordered dict result in changes in chunk options")
@pytest.mark.parametrize('script_file', list_all_notebooks('.jl'))
def test_julia_unchanged(script_file):
    with open(script_file, encoding='utf-8') as fp:
        julia = fp.read()

    ipynb_file = mirror_file(script_file)
    create_if_missing(ipynb_file, script_file)

    julia_ref = jupytext.writes(jupytext.readf(ipynb_file), ext='.jl')
    compare(julia, julia_ref)



"""Read and write notebooks as RStudio notebook files, with .Rmd extension.

Raw and markdown cells are converted to markdown, while code cells are
converted to code chunks. The transformation is reversible and all inputs
are preserved (not outputs, though).

Authors:

* Marc Wouts
"""

import os
import io
from copy import deepcopy
from nbformat.v4.rwbase import NotebookReader, NotebookWriter
from nbformat.v4.nbbase import new_notebook
import nbformat

from .header import header_to_metadata_and_cell, metadata_and_cell_to_header, \
    encoding_and_executable
from .languages import default_language_from_metadata_and_ext, \
    set_main_and_cell_language
from .cell_to_text import CellExporter
from .cell_reader import CellReader

NOTEBOOK_EXTENSIONS = ['.ipynb', '.Rmd', '.md', '.jl', '.py', '.R']


def markdown_comment(ext):
    """Markdown escape for given notebook extension"""
    return '' if ext in ['.Rmd', '.md'] else "#'" if ext == '.R' else "#"


class TextNotebookReader(NotebookReader):
    """Text notebook reader"""

    def __init__(self, ext):
        self.ext = ext
        self.prefix = markdown_comment(ext)

    header_to_metadata_and_cell = header_to_metadata_and_cell

    def markdown_unescape(self, line):
        """Remove markdown escape, if any"""
        if self.prefix == '':
            return line
        line = line[len(self.prefix):]
        if line.startswith(' '):
            line = line[1:]
        return line

    def reads(self, s, **kwargs):
        """Read a notebook from text"""
        return self.to_notebook(s)

    def to_notebook(self, text):
        """Read a notebook from text"""
        lines = text.splitlines()

        cells = []
        metadata, header_cell, pos = \
            self.header_to_metadata_and_cell(lines)

        if header_cell:
            cells.append(header_cell)

        if pos > 0:
            lines = lines[pos:]

        while lines:
            reader = CellReader(self.ext)
            cell, pos = reader.read(lines)
            cells.append(cell)
            if pos <= 0:
                raise Exception('Blocked at lines ' + '\n'.join(lines[:6]))
            lines = lines[pos:]

        set_main_and_cell_language(metadata, cells, self.ext)

        # Backward compatibility with nbrmd
        for key in ['nbrmd_formats', 'nbrmd_format_version']:
            if key in metadata:
                metadata[key.replace('nbrmd', 'jupytext')] = \
                    metadata.pop(key)

        return new_notebook(cells=cells, metadata=metadata)


class TextNotebookWriter(NotebookWriter):
    """Write notebook to their text representations"""

    def __init__(self, ext):
        self.ext = ext
        self.prefix = markdown_comment(ext)

    def markdown_escape(self, lines):
        """
        Escape markdown text
        :param lines:
        :return:
        """
        if self.prefix == '':
            return lines
        return [self.prefix if line == '' else self.prefix + ' ' + line
                for line in lines]

    encoding_and_executable = encoding_and_executable
    metadata_and_cell_to_header = metadata_and_cell_to_header

    def writes(self, nb, **kwargs):
        """Write the text representation of a notebook to a string"""
        nb = deepcopy(nb)
        default_language = default_language_from_metadata_and_ext(nb, self.ext)
        if 'main_language' in nb.metadata:
            del nb.metadata['main_language']

        lines = self.encoding_and_executable(nb)
        lines.extend(self.metadata_and_cell_to_header(nb))

        cells = [CellExporter(cell, default_language, self.ext)
                 for cell in nb.cells]

        texts = [cell.cell_to_text() for cell in cells]

        for i, cell in enumerate(cells):
            text = texts[i]

            # Simplify cell marker when previous line is blank
            if text[0] == '# + {}' and (not lines or not lines[-1]):
                text[0] = '# +'

            # remove end of cell marker when redundant
            # with next explicit marker
            if self.ext in ['.py', '.jl'] and cell.is_code() \
                    and text[-1] == '# -':
                if cell.lines_to_end_of_cell_marker:
                    text = text[:-1] + \
                           [''] * cell.lines_to_end_of_cell_marker + ['# -']
                elif i + 1 >= len(texts) or \
                        (texts[i + 1][0].startswith('# + {')):
                    text = text[:-1]

            lines.extend(text)
            lines.extend([''] * cell.lines_to_next_cell)

            # two blank lines between markdown cells in Rmd
            if self.ext in ['.Rmd', '.md'] and not cell.is_code():
                if i + 1 < len(cells) and not cells[i + 1].is_code():
                    lines.append('')

        return '\n'.join(lines)


_NOTEBOOK_READERS = {ext: TextNotebookReader(ext)
                     for ext in NOTEBOOK_EXTENSIONS if ext != '.ipynb'}
_NOTEBOOK_WRITERS = {ext: TextNotebookWriter(ext)
                     for ext in NOTEBOOK_EXTENSIONS if ext != '.ipynb'}


def reads(text, ext, as_version=4, **kwargs):
    """Read a notebook from a string"""
    if ext == '.ipynb':
        return nbformat.reads(text, as_version, **kwargs)

    return _NOTEBOOK_READERS[ext].reads(text, **kwargs)


def read(file_or_stream, ext, as_version=4, **kwargs):
    """Read a notebook from a file"""
    if ext == '.ipynb':
        return nbformat.read(file_or_stream, as_version, **kwargs)

    return _NOTEBOOK_READERS[ext].read(file_or_stream, **kwargs)


def writes(notebook, ext, version=nbformat.NO_CONVERT, **kwargs):
    """Write a notebook to a string"""
    if ext == '.ipynb':
        return nbformat.writes(notebook, version, **kwargs)

    return _NOTEBOOK_WRITERS[ext].writes(notebook)


def write(notebook, file_or_stream, ext, version=nbformat.NO_CONVERT,
          **kwargs):
    """Write a notebook to a file"""
    if ext == '.ipynb':
        return nbformat.write(notebook, file_or_stream, version, **kwargs)

    return _NOTEBOOK_WRITERS[ext].write(notebook, file_or_stream)


def readf(nb_file):
    """Read a notebook from the file with given name"""
    _, ext = os.path.splitext(nb_file)
    if ext not in NOTEBOOK_EXTENSIONS:
        raise TypeError(
            'File {} is not a notebook. '
            'Expected extensions are {}'.format(nb_file,
                                                NOTEBOOK_EXTENSIONS))
    with io.open(nb_file, encoding='utf-8') as stream:
        return read(stream, as_version=4, ext=ext)


def writef(notebook, nb_file):
    """Write a notebook to the file with given name"""
    _, ext = os.path.splitext(nb_file)
    if ext not in NOTEBOOK_EXTENSIONS:
        raise TypeError(
            'File {} is not a notebook. '
            'Expected extensions are {}'.format(nb_file,
                                                NOTEBOOK_EXTENSIONS))
    with io.open(nb_file, 'w', encoding='utf-8') as stream:
        write(notebook, stream, version=nbformat.NO_CONVERT, ext=ext)



import os
import copy


def list_all_notebooks(ext, path=None):
    """
    :ext: desired extension
    :return: all notebooks in the directory of this script,
     with the desired extension
    """
    nb_path = os.path.dirname(os.path.abspath(__file__))
    if path:
        nb_path = os.path.join(nb_path, path)
    notebooks = []
    for nb_file in os.listdir(nb_path):
        _, nb_ext = os.path.splitext(nb_file)
        if nb_ext.lower() == ext.lower() and \
                (not nb_file.startswith('ir_notebook')
                 or nb_file.startswith('R_sample')):
            notebooks.append(os.path.join(nb_path, nb_file))
    return notebooks


def list_r_notebooks(ext):
    """
    :ext: desired extension
    :return: all R notebooks in the directory of this script,
     with the desired extension
    """
    nb_path = os.path.dirname(os.path.abspath(__file__))
    notebooks = []
    for nb_file in os.listdir(nb_path):
        _, nb_ext = os.path.splitext(nb_file)
        if nb_ext.lower() == ext.lower() and \
                (nb_file.startswith('ir_notebook')
                 or nb_file.startswith('R_sample')):
            notebooks.append(os.path.join(nb_path, nb_file))
    return notebooks


def list_julia_notebooks(ext):
    """
    :ext: desired extension
    :return: all Julia notebooks in the directory of this script,
     with the desired extension
    """
    nb_path = os.path.dirname(os.path.abspath(__file__))
    notebooks = []
    for nb_file in os.listdir(nb_path):
        _, nb_ext = os.path.splitext(nb_file)
        if nb_ext.lower() == ext.lower() and nb_file.startswith('julia_'):
            notebooks.append(os.path.join(nb_path, nb_file))
    return notebooks


def list_py_notebooks(ext):
    """
    :ext: desired extension
    :return: all Python notebooks in the directory of this script,
     with the desired extension
    """
    nb_path = os.path.dirname(os.path.abspath(__file__))
    notebooks = []
    for nb_file in os.listdir(nb_path):
        _, nb_ext = os.path.splitext(nb_file)
        if nb_ext.lower() == ext.lower() and not (
                nb_file.startswith('julia_')
                or nb_file.startswith('ir_notebook')
                or nb_file.startswith('R_sample')):
            notebooks.append(os.path.join(nb_path, nb_file))
    return notebooks



import os
import sys
import subprocess
import pytest
import mock
import jupytext
from .utils import list_all_notebooks, list_r_notebooks, \
    list_py_notebooks, list_julia_notebooks


@pytest.mark.skipif(sys.version_info < (3, 6),
                    reason="unordered dict result in changes in chunk options")
@pytest.mark.skipif(isinstance(jupytext.RMarkdownExporter, str),
                    reason=jupytext.RMarkdownExporter)
@pytest.mark.parametrize('nb_file', list_all_notebooks('.ipynb'))
def test_nbconvert_and_read(nb_file):
    # Load notebook
    nb = jupytext.readf(nb_file)

    # Export to Rmd using jupytext package
    rmd1 = jupytext.writes(nb, ext='.Rmd')

    # Export to Rmd using nbconvert exporter
    rmd_exporter = jupytext.RMarkdownExporter()
    (rmd2, resources) = rmd_exporter.from_notebook_node(nb)

    assert rmd1 == rmd2


@pytest.mark.skipif(isinstance(jupytext.PyNotebookExporter, str),
                    reason=jupytext.PyNotebookExporter)
@pytest.mark.parametrize('nb_file', list_py_notebooks('.ipynb'))
def test_nbconvert_and_read_py(nb_file):
    # Load notebook
    nb = jupytext.readf(nb_file)

    # Export to py using jupytext package
    py1 = jupytext.writes(nb, ext='.py')

    # Export to py using nbconvert exporter
    py_exporter = jupytext.PyNotebookExporter()
    (py2, resources) = py_exporter.from_notebook_node(nb)

    assert py1 == py2


@pytest.mark.skipif(isinstance(jupytext.JlNotebookExporter, str),
                    reason=jupytext.JlNotebookExporter)
@pytest.mark.parametrize('nb_file', list_julia_notebooks('.ipynb'))
def test_nbconvert_and_read_julia(nb_file):
    # Load notebook
    nb = jupytext.readf(nb_file)

    # Export to py using jupytext package
    julia = jupytext.writes(nb, ext='.jl')

    # Export to py using nbconvert exporter
    julia_exporter = jupytext.JlNotebookExporter()
    (julia2, resources) = julia_exporter.from_notebook_node(nb)

    assert julia == julia2


@pytest.mark.skipif(isinstance(jupytext.RNotebookExporter, str),
                    reason=jupytext.RNotebookExporter)
@pytest.mark.parametrize('nb_file', list_r_notebooks('.ipynb'))
def test_nbconvert_and_read_r(nb_file):
    # Load notebook
    nb = jupytext.readf(nb_file)

    # Export to py using jupytext package
    r1 = jupytext.writes(nb, ext='.R')

    # Export to py using nbconvert exporter
    r_exporter = jupytext.RNotebookExporter()
    (r2, resources) = r_exporter.from_notebook_node(nb)

    assert r1 == r2


pytest.importorskip('jupyter')


@pytest.mark.skipif(sys.version_info < (3, 6),
                    reason="unordered dict result in changes in chunk options")
@pytest.mark.skipif(isinstance(jupytext.RMarkdownExporter, str),
                    reason=jupytext.RMarkdownExporter)
@pytest.mark.parametrize('nb_file', list_py_notebooks('.ipynb'))
def test_nbconvert_cmd_line(nb_file, tmpdir):
    rmd_file = str(tmpdir.join('notebook.Rmd'))

    subprocess.call(['jupyter', 'nbconvert', '--to', 'rmarkdown',
                     nb_file, '--output', rmd_file])

    assert os.path.isfile(rmd_file)

    nb = jupytext.readf(nb_file)
    with mock.patch('jupytext.file_format_version.FILE_FORMAT_VERSION',
                    jupytext.file_format_version.FILE_FORMAT_VERSION_ORG):
        rmd1 = jupytext.writes(nb, ext='.Rmd')
    with open(rmd_file) as fp:
        rmd2 = fp.read()

    assert rmd1 == rmd2


@pytest.mark.skipif(sys.version_info < (3, 6),
                    reason="unordered dict result in changes in chunk options")
@pytest.mark.skipif(isinstance(jupytext.PyNotebookExporter, str),
                    reason=jupytext.PyNotebookExporter)
@pytest.mark.parametrize('nb_file', list_py_notebooks('.ipynb'))
def test_nbconvert_cmd_line_py(nb_file, tmpdir):
    py_file = str(tmpdir.join('notebook.py'))

    subprocess.call(['jupyter', 'nbconvert', '--to', 'pynotebook',
                     nb_file, '--output', py_file])

    assert os.path.isfile(py_file)

    nb = jupytext.readf(nb_file)
    with mock.patch('jupytext.file_format_version.FILE_FORMAT_VERSION',
                    jupytext.file_format_version.FILE_FORMAT_VERSION_ORG):
        py1 = jupytext.writes(nb, ext='.py')
    with open(py_file) as fp:
        py2 = fp.read()

    assert py1 == py2


@pytest.mark.skipif(sys.version_info < (3, 6),
                    reason="unordered dict result in changes in chunk options")
@pytest.mark.skipif(isinstance(jupytext.RNotebookExporter, str),
                    reason=jupytext.RNotebookExporter)
@pytest.mark.parametrize('nb_file', list_r_notebooks('.ipynb'))
def test_nbconvert_cmd_line_R(nb_file, tmpdir):
    r_file = str(tmpdir.join('notebook.R'))

    subprocess.call(['jupyter', 'nbconvert', '--to', 'rnotebook',
                     nb_file, '--output', r_file])

    assert os.path.isfile(r_file)

    nb = jupytext.readf(nb_file)
    with mock.patch('jupytext.file_format_version.FILE_FORMAT_VERSION',
                    jupytext.file_format_version.FILE_FORMAT_VERSION_ORG):
        r = jupytext.writes(nb, ext='.R')
    with open(r_file) as fp:
        r2 = fp.read()

    assert r == r2



"""Current file format for the various text only notebook extensions"""

import os

FILE_FORMAT_VERSION = {
    # R markdown format
    '.Rmd': '1.0',
    # Version 1.0 on 2018-08-22 - jupytext v0.5.2 : Initial version

    # Markdown format
    '.md': '1.0',
    # Version 1.0 on 2018-08-31 - jupytext v0.6.0 : Initial version

    # Python and Julia scripts
    '.py': '1.2',
    '.jl': '1.2',
    # Version 1.2 on 2018-09-05 - jupytext v0.6.3 : Metadata bracket can be
    # omitted when empty, if previous line is empty #57
    # Version 1.1 on 2018-08-25 - jupytext v0.6.0 : Cells separated with one
    # blank line #38
    # Version 1.0 on 2018-08-22 - jupytext v0.5.2 : Initial version

    # Python scripts
    '.R': '1.0',
    # Version 1.0 on 2018-08-22 - jupytext v0.5.2 : Initial version
}

MIN_FILE_FORMAT_VERSION = {'.Rmd': '1.0', '.md': '1.0', '.py': '1.1',
                           '.jl': '1.1', '.R': '1.0'}

FILE_FORMAT_VERSION_ORG = FILE_FORMAT_VERSION


def file_format_version(ext):
    """Return file format version for given ext"""
    return FILE_FORMAT_VERSION.get(ext)


def min_file_format_version(ext):
    """Return minimum file format version for given ext"""
    return MIN_FILE_FORMAT_VERSION.get(ext)


def check_file_version(notebook, source_path, outputs_path):
    """Raise if file version in source file would override outputs"""
    _, ext = os.path.splitext(source_path)
    current = file_format_version(ext)
    version = notebook.metadata.get('jupytext_format_version')
    if version:
        del notebook.metadata['jupytext_format_version']

    # Missing version, still generated by jupytext?
    if notebook.metadata and not version:
        version = current

    # Same version? OK
    if version == current:
        return

    # Version larger than minimum readable version
    if version <= current and version >= min_file_format_version(ext):
        return

        # Not merging? OK
    if source_path == outputs_path:
        return

    raise ValueError("File {} has jupytext_format_version={}, but "
                     "current version for that extension is {}.\n"
                     "It would not be safe to override the source of {} "
                     "with that file.\n"
                     "Please remove one or the other file."
                     .format(os.path.basename(source_path),
                             version, current,
                             os.path.basename(outputs_path)))



import sys
import pytest
import jupytext
from .utils import list_all_notebooks

jupytext.file_format_version.FILE_FORMAT_VERSION = {}


@pytest.mark.skipif(sys.version_info < (3, 6),
                    reason="unordered dict result in changes in chunk options")
@pytest.mark.parametrize('r_file',
                         list_all_notebooks('.R'))
def test_jupytext_same_as_knitr_spin(r_file, tmpdir):
    nb = jupytext.readf(r_file)
    rmd_jupytext = jupytext.writes(nb, ext='.Rmd')

    # Rmd file generated with spin(hair='R/spin.R', knit=FALSE)
    rmd_file = r_file.replace('.R', '.Rmd')

    with open(rmd_file) as fp:
        rmd_spin = fp.read()

    assert rmd_spin == rmd_jupytext



"""ContentsManager that allows to open Rmd, py, R and ipynb files as notebooks
"""
import os
import nbformat
import mock
import six

try:
    import notebook.transutils  # noqa
except ImportError:
    pass
from notebook.services.contents.filemanager import FileContentsManager
from tornado.web import HTTPError
from traitlets import Unicode
from traitlets.config import Configurable
import jupytext

from . import combine
from .file_format_version import check_file_version


def _jupytext_writes(ext):
    def _writes(nbk, version=nbformat.NO_CONVERT, **kwargs):
        return jupytext.writes(nbk, version=version, ext=ext, **kwargs)

    return _writes


def _jupytext_reads(ext):
    def _reads(text, as_version, **kwargs):
        return jupytext.reads(text, ext, as_version, **kwargs)

    return _reads


def check_formats(formats):
    """
    Parse, validate and return notebooks extensions
    :param formats: a list of lists of notebook extensions,
    or a colon separated string of extension groups, comma separated
    :return: list of lists (groups) of notebook extensions
    """

    # Parse formats represented as strings
    if not isinstance(formats, list):
        formats = [group.split(',') for group in formats.split(';')]

    expected_format = ("Notebook metadata 'jupytext_formats' should "
                       "be a list of extension groups, like 'ipynb,Rmd'.\n"
                       "Groups can be separated with colon, for instance: "
                       "'ipynb,nb.py;script.ipynb,py'")

    validated_formats = []
    for group in formats:
        if not isinstance(group, list):
            # In early versions (0.4 and below), formats could be a list of
            # extensions. We understand this as a single group
            return check_formats([formats])
        validated_group = []
        for fmt in group:
            if not isinstance(fmt, six.string_types):
                raise ValueError('Extensions should be strings among {}'
                                 ', not {}.\n{}'
                                 .format(str(jupytext.NOTEBOOK_EXTENSIONS),
                                         str(fmt),
                                         expected_format))
            if fmt == '':
                continue
            if not fmt.startswith('.'):
                fmt = '.' + fmt
            if not any([fmt.endswith(ext)
                        for ext in jupytext.NOTEBOOK_EXTENSIONS]):
                raise ValueError('Group extension {} contains {}, '
                                 'which does not end with either {}.\n{}'
                                 .format(str(group), fmt,
                                         str(jupytext.NOTEBOOK_EXTENSIONS),
                                         expected_format))
            validated_group.append(fmt)

        if validated_group:
            validated_formats.append(validated_group)

    return validated_formats


def file_fmt_ext(path):
    """
    Return file name, format (possibly .nb.py) and extension (.py)
    """
    file, ext = os.path.splitext(path)
    file, intermediate_ext = os.path.splitext(file)
    return file, intermediate_ext + ext, ext


class TextFileContentsManager(FileContentsManager, Configurable):
    """
    A FileContentsManager Class that reads and stores notebooks to classical
    Jupyter notebooks (.ipynb), R Markdown notebooks (.Rmd), Julia (.jl),
    Python (.py) or R scripts (.R)
    """

    nb_extensions = [ext for ext in jupytext.NOTEBOOK_EXTENSIONS if
                     ext != '.ipynb']

    def all_nb_extensions(self):
        """
        Notebook extensions, including ipynb
        :return:
        """
        return ['.ipynb'] + self.nb_extensions

    default_jupytext_formats = Unicode(
        u'',
        help='Save notebooks to these file extensions. '
             'Can be any of ipynb,Rmd,md,jl,py,R,nb.jl,nb.py,nb.R '
             'comma separated',
        config=True)

    def format_group(self, fmt, nbk=None):
        """Return the group of extensions that contains 'fmt'"""
        jupytext_formats = ((nbk.metadata.get('jupytext_formats')
                             if nbk else None) or
                            self.default_jupytext_formats)

        try:
            jupytext_formats = check_formats(jupytext_formats)
        except ValueError as err:
            raise HTTPError(400, str(err))

        # Find group that contains the current format
        for group in jupytext_formats:
            if fmt in group:
                return group

        if ['.ipynb'] in jupytext_formats:
            return [fmt, '.ipynb']

        return [fmt]

    def _read_notebook(self, os_path, as_version=4,
                       load_alternative_format=True):
        """Read a notebook from an os path."""
        file, fmt, ext = file_fmt_ext(os_path)
        if ext in self.nb_extensions:
            with mock.patch('nbformat.reads', _jupytext_reads(ext)):
                nbk = super(TextFileContentsManager, self) \
                    ._read_notebook(os_path, as_version)
        else:
            nbk = super(TextFileContentsManager, self) \
                ._read_notebook(os_path, as_version)

        if not load_alternative_format:
            return nbk

        fmt_group = self.format_group(fmt, nbk)

        source_format = fmt
        outputs_format = fmt

        # Source format is first non ipynb format found on disk
        if fmt.endswith('.ipynb'):
            for alt_fmt in fmt_group:
                if not alt_fmt.endswith('.ipynb') and \
                        os.path.isfile(file + alt_fmt):
                    source_format = alt_fmt
                    break
        # Outputs taken from ipynb if in group, if file exists
        else:
            for alt_fmt in fmt_group:
                if alt_fmt.endswith('.ipynb') and \
                        os.path.isfile(file + alt_fmt):
                    outputs_format = alt_fmt
                    break

        if source_format != fmt:
            self.log.info('Reading SOURCE from {}'
                          .format(os.path.basename(file + source_format)))
            nb_outputs = nbk
            nbk = self._read_notebook(file + source_format,
                                      as_version=as_version,
                                      load_alternative_format=False)
        elif outputs_format != fmt:
            self.log.info('Reading OUTPUTS from {}'
                          .format(os.path.basename(file + outputs_format)))
            if outputs_format != fmt:
                nb_outputs = self._read_notebook(file + outputs_format,
                                                 as_version=as_version,
                                                 load_alternative_format=False)
        else:
            nb_outputs = None

        try:
            check_file_version(nbk, file + source_format,
                               file + outputs_format)
        except ValueError as err:
            raise HTTPError(400, str(err))

        if nb_outputs:
            combine.combine_inputs_with_outputs(nbk, nb_outputs)
            if self.notary.check_signature(nb_outputs):
                self.notary.sign(nbk)
        elif not fmt.endswith('.ipynb'):
            self.notary.sign(nbk)

        return nbk

    def _save_notebook(self, os_path, nb):
        """Save a notebook to an os_path."""
        os_file, fmt, _ = file_fmt_ext(os_path)
        for alt_fmt in self.format_group(fmt, nb):
            os_path_fmt = os_file + alt_fmt
            self.log.info("Saving %s", os.path.basename(os_path_fmt))
            alt_ext = '.' + alt_fmt.split('.')[-1]
            if alt_ext in self.nb_extensions:
                with mock.patch('nbformat.writes', _jupytext_writes(alt_ext)):
                    super(TextFileContentsManager, self) \
                        ._save_notebook(os_path_fmt, nb)
            else:
                super(TextFileContentsManager, self) \
                    ._save_notebook(os_path_fmt, nb)

    def get(self, path, content=True, type=None, format=None):
        """ Takes a path for an entity and returns its model"""
        path = path.strip('/')

        if self.exists(path) and \
                (type == 'notebook' or
                 (type is None and
                  any([path.endswith(ext)
                       for ext in self.all_nb_extensions()]))):
            return self._notebook_model(path, content=content)

        return super(TextFileContentsManager, self) \
            .get(path, content, type, format)

    def trust_notebook(self, path):
        """Trust the current notebook"""
        file, fmt, _ = file_fmt_ext(path)
        for alt_fmt in self.format_group(fmt):
            if alt_fmt.endswith('.ipynb'):
                super(TextFileContentsManager, self) \
                    .trust_notebook(file + alt_fmt)

    def rename_file(self, old_path, new_path):
        """Rename the current notebook, as well as its
         alternative representations"""
        old_file, org_fmt, _ = file_fmt_ext(old_path)
        new_file, new_fmt, _ = file_fmt_ext(new_path)

        if org_fmt == new_fmt:
            for alt_fmt in self.format_group(org_fmt):
                if self.file_exists(old_file + alt_fmt):
                    super(TextFileContentsManager, self) \
                        .rename_file(old_file + alt_fmt, new_file + alt_fmt)
        else:
            super(TextFileContentsManager, self) \
                .rename_file(old_path, new_path)



"""
R and Py notebook exporters for nbconvert
"""

from traitlets import default
from nbconvert.exporters import Exporter
import jupytext


class PyNotebookExporter(Exporter):
    """
    Exports to a python notebook (.py)
    """

    @default('file_extension')
    def _file_extension_default(self):
        return '.py'

    def from_notebook_node(self, nb, resources=None, **kw):
        resources = resources or {}
        resources['output_extension'] = self.file_extension
        return jupytext.writes(nb, ext='.py'), resources


class JlNotebookExporter(Exporter):
    """
    Exports to a julia notebook (.jl)
    """

    @default('file_extension')
    def _file_extension_default(self):
        return '.jl'

    def from_notebook_node(self, nb, resources=None, **kw):
        resources = resources or {}
        resources['output_extension'] = self.file_extension
        return jupytext.writes(nb, ext='.jl'), resources


class RNotebookExporter(Exporter):
    """
    Exports to a R notebook (.R)
    """

    @default('file_extension')
    def _file_extension_default(self):
        return '.R'

    def from_notebook_node(self, nb, resources=None, **kw):
        resources = resources or {}
        resources['output_extension'] = self.file_extension
        return jupytext.writes(nb, ext='.R'), resources



"""R markdown notebook format for Jupyter

Use this module to read or write Jupyter notebooks as Markdown documents,
R markdown documents, Python or R scripts (methods 'read', 'reads', 'write',
'writes')

Use the TextFileContentsManager to open Rmd and Jupyter notebooks in Jupyter

Use the 'jupytext' conversion script to convert Jupyter notebooks from/to
R Markdown notebooks.
"""

from .jupytext import readf, writef, writes, reads, NOTEBOOK_EXTENSIONS
from .file_format_version import FILE_FORMAT_VERSION

try:
    from .rmarkdownexporter import RMarkdownExporter
    from .srcexporter import PyNotebookExporter
    from .srcexporter import RNotebookExporter
    from .srcexporter import JlNotebookExporter
except ImportError as err:
    RMarkdownExporter = PyNotebookExporter = JlNotebookExporter =\
        RNotebookExporter = str(err)

try:
    from .contentsmanager import TextFileContentsManager
except ImportError as err:
    TextFileContentsManager = str(err)

__all__ = ['readf', 'writef', 'writes', 'reads',
           'NOTEBOOK_EXTENSIONS', 'FILE_FORMAT_VERSION',
           'RMarkdownExporter', 'PyNotebookExporter', 'RNotebookExporter',
           'TextFileContentsManager']



"""Command line conversion tools `jupytext` and `nbsrc`
"""

import os
import argparse
from jupytext import readf, writef, writes
from jupytext import NOTEBOOK_EXTENSIONS
from .combine import combine_inputs_with_outputs
from .compare import test_round_trip_conversion
from .file_format_version import check_file_version


def convert_notebook_files(nb_files, ext, output=None,
                           test_round_trip=False, preserve_outputs=True):
    """
    Export R markdown notebooks, python or R scripts, or Jupyter notebooks,
    to the opposite format
    :param nb_files: one or more notebooks files
    :param ext: extension of destination
    :param output: None, destination file, or '-' for stdout
    :param test_round_trip: should round trip conversion be tested?
    :param preserve_outputs: preserve the current outputs of .ipynb file
    when possible
    :return:
    """

    if len(nb_files) > 1 and output:
        raise ValueError(
            "Output argument can only be used with a single notebook")

    for nb_file in nb_files:
        dest, current_ext = os.path.splitext(nb_file)
        if current_ext not in NOTEBOOK_EXTENSIONS:
            raise TypeError('File {} is not a notebook'.format(nb_file))

        notebook = readf(nb_file)
        if test_round_trip:
            test_round_trip_conversion(notebook, ext, preserve_outputs)
            continue

        if output == '-':
            print(writes(notebook, ext=ext))
            continue

        if output:
            dest, dest_ext = os.path.splitext(output)
            if dest_ext != ext:
                raise TypeError('Destination extension {} is not consistent'
                                'with format {} '.format(dest_ext, ext))

        save_notebook_as(notebook, nb_file, dest + ext, preserve_outputs)


def save_notebook_as(notebook, nb_file, nb_dest, combine):
    """Save notebook to file, in desired format"""
    if combine and os.path.isfile(nb_dest) and \
            os.path.splitext(nb_dest)[1] == '.ipynb':
        check_file_version(notebook, nb_file, nb_dest)
        nb_outputs = readf(nb_dest)
        combine_inputs_with_outputs(notebook, nb_outputs)

    writef(notebook, nb_dest)


def cli_jupytext(args=None):
    """Command line parser for jupytext"""
    parser = argparse.ArgumentParser(
        description='Jupyter notebooks as markdown documents, '
                    'Julia, Python or R scripts')

    parser.add_argument(
        '--to',
        choices=['notebook', 'python', 'julia', 'r', 'markdown',
                 'rmarkdown'] + [ext.replace('.', '') for ext in
                                 NOTEBOOK_EXTENSIONS],
        help="Destination format")
    parser.add_argument('notebooks',
                        help='One or more notebook(s) to be converted',
                        nargs='+')
    parser.add_argument('-o', '--output',
                        help='Destination file. Defaults to original file, '
                             'with extension changed to destination format. '
                             "Use '-' for printing the notebook on stdout.")
    parser.add_argument('--update', action='store_true',
                        help='Preserve outputs of .ipynb destination '
                             '(when file exists and inputs match)')
    parser.add_argument('--test', dest='test', action='store_true',
                        help='Test that notebook is stable under '
                             'round trip conversion')
    args = parser.parse_args(args)

    if not args.to:
        if args.output and args.output != '-':
            _, ext = os.path.splitext(args.output)
        else:
            raise TypeError('Please specificy either --to or --output')
    elif args.to in NOTEBOOK_EXTENSIONS:
        pass
    elif '.' + args.to in NOTEBOOK_EXTENSIONS:
        ext = '.' + args.to
    else:
        ext = {'notebook': '.ipynb',
               'python': '.py',
               'julia': '.jl',
               'r': '.R',
               'markdown': '.md',
               'rmarkdown': '.Rmd'}[args.to]

    if ext not in NOTEBOOK_EXTENSIONS:
        raise TypeError('Output extensions should be in {}'
                        .format(", ".join(NOTEBOOK_EXTENSIONS)))

    if args.update and ext != '.ipynb':
        raise ValueError('--update works exclusively with --to notebook')

    args.to = ext

    return args


def jupytext(args=None):
    """Entry point for the jupytext script"""
    try:
        args = cli_jupytext(args)
        convert_notebook_files(nb_files=args.notebooks,
                               ext=args.to,
                               output=args.output,
                               test_round_trip=args.test,
                               preserve_outputs=args.update)
    except (ValueError, TypeError, IOError) as err:
        print('jupytext: error: ' + str(err))
        exit(1)



"""Export notebook cells as text"""

import re
from copy import copy
from .languages import cell_language
from .cell_reader import CellReader
from .cell_metadata import filter_metadata, is_active, \
    metadata_to_rmd_options, metadata_to_json_options
from .magics import escape_magic, escape_code_start


def cell_source(cell):
    """
    Return the source of the current cell, as an array of lines
    :param cell:
    :return:
    """
    source = cell.source
    if source == '':
        return ['']
    if source.endswith('\n'):
        return source.splitlines() + ['']
    return source.splitlines()


def code_to_rmd(source, metadata, language):
    """
    Represent a code cell with given source and metadata as a rmd cell
    :param source:
    :param metadata:
    :param language:
    :return:
    """
    lines = []
    if not is_active('Rmd', metadata):
        metadata['eval'] = False
    options = metadata_to_rmd_options(language, metadata)
    lines.append('```{{{}}}'.format(options))
    lines.extend(source)
    lines.append('```')
    return lines


def code_to_md(source, metadata, language):
    """
    Represent a code cell with given source and metadata as a md cell
    :param source:
    :param metadata:
    :param language:
    :return:
    """
    options = []
    if language:
        options.append(language)
    if 'name' in metadata:
        options.append(metadata['name'])

    return ['```{}'.format(' '.join(options))] + source + ['```']


def code_to_r(source, metadata):
    """
    Represent a code cell with given source and metadata as a R cell
    :param source:
    :param metadata:
    :return:
    """
    lines = []
    if not is_active('R', metadata):
        metadata['eval'] = False
    options = metadata_to_rmd_options(None, metadata)
    if options:
        lines.append('#+ {}'.format(options))
    lines.extend(source)
    return lines


def code_to_py(source, metadata):
    """
    Represent a code cell with given source and metadata as a python cell
    """
    lines = []
    if not metadata:
        return source

    endofcell = metadata['endofcell']
    if endofcell == '-':
        del metadata['endofcell']
    options = metadata_to_json_options(metadata)
    lines.append('# + {}'.format(options))
    lines.extend(source)
    lines.append('# {}'.format(endofcell))
    return lines


def py_endofcell_marker(source):
    """Issues #31 #38:  does the cell contain a blank line? In that case
    we add an end-of-cell marker"""
    endofcell = '-'
    while True:
        endofcell_re = re.compile(r'^#( )' + endofcell + r'\s*$')
        if list(filter(endofcell_re.match, source)):
            endofcell = endofcell + '-'
        else:
            return endofcell


class CellExporter():
    """A class that represent a notebook cell as text"""

    def __init__(self, cell, default_language, ext):
        self.ext = ext
        self.prefix = '' if ext in ['.Rmd', '.md'] \
            else "#'" if ext == '.R' else '#'
        self.cell_type = cell.cell_type
        self.source = cell_source(cell)
        self.metadata = filter_metadata(cell.metadata)
        self.language = cell_language(self.source) or default_language

        # how many blank lines before next cell
        self.lines_to_next_cell = cell.metadata.get('lines_to_next_cell', 1)
        self.lines_to_end_of_cell_marker = \
            cell.metadata.get('lines_to_end_of_cell_marker', 0)

        # for compatibility with v0.5.4 and lower (to be removed)
        if 'skipline' in cell.metadata:
            self.lines_to_next_cell += 1
        if 'noskipline' in cell.metadata:
            self.lines_to_next_cell -= 1

        if cell.cell_type == 'raw' and 'active' not in self.metadata:
            self.metadata['active'] = ''

    def is_code(self):
        """Is this cell a code cell?"""
        if self.cell_type == 'code':
            return True
        if self.cell_type == 'raw' and 'active' in self.metadata:
            return True
        return False

    def cell_to_text(self):
        """Return the text representation for the cell"""
        if self.is_code():
            return self.code_to_text()

        source = copy(self.source)
        if self.ext in ['.Rmd', '.md']:
            escape_code_start(source, self.ext, None)
        return self.markdown_escape(source)

    def markdown_escape(self, source):
        """Escape the given source, for a markdown cell"""
        if self.ext in ['.Rmd', '.md']:
            return source
        return [self.prefix + ' ' + line if line else self.prefix
                for line in source]

    def explicit_start_marker(self, source):
        """Does the python representation of this cell requires an explicit
        start of cell marker?"""
        if self.metadata:
            return True
        if all([line.startswith('#') for line in self.source]):
            return True
        if CellReader(self.ext).read(source)[1] < len(source):
            return True

        return False

    def code_to_text(self):
        """Return the text representation of a code cell"""
        active = is_active(self.ext, self.metadata)
        if self.ext in ['.R', '.py', '.jl']:
            if active and self.language != (
                    'R' if self.ext == '.R' else
                    'python' if self.ext == '.py' else 'julia'):
                active = False
                self.metadata['active'] = 'ipynb'
                self.metadata['language'] = self.language

        source = copy(self.source)
        escape_code_start(source, self.ext, self.language)

        if active and self.ext != '.md':
            escape_magic(source, self.language)

        if self.ext == '.Rmd':
            return code_to_rmd(source, self.metadata, self.language)

        if self.ext == '.md':
            return code_to_md(
                source, self.metadata,
                self.language if self.cell_type == 'code' else None)

        if not active:
            source = ['# ' + line if line else '#' for line in source]

        if self.ext == '.R':
            return code_to_r(source, self.metadata)

        # py, jl
        if self.explicit_start_marker(source):
            self.metadata['endofcell'] = py_endofcell_marker(source)

        return code_to_py(source, self.metadata)



"""Read notebook cells from their text representation"""

import re
from nbformat.v4.nbbase import new_code_cell, new_raw_cell, new_markdown_cell
from .cell_metadata import is_active, json_options_to_metadata, \
    md_options_to_metadata, rmd_options_to_metadata
from .stringparser import StringParser
from .magics import unescape_magic, is_magic, unescape_code_start

_START_CODE_RMD = re.compile(r"^```{(.*)}\s*$")
_START_CODE_MD = re.compile(r"^```(.*)")
_END_CODE_MD = re.compile(r"^```\s*$")
_CODE_OPTION_R = re.compile(r"^#\+(.*)\s*$")
_CODE_OPTION_PY = re.compile(r"^(#|# )\+(\s*){(.*)}\s*$")
_SIMPLE_START_CODE_PY = re.compile(r"^(#|# )\+(\s*)$")
_BLANK_LINE = re.compile(r"^\s*$")
_PY_COMMENT = re.compile(r"^\s*#")
_PY_INDENTED = re.compile(r"^\s")


def uncomment(lines):
    """Remove quote and space when possible"""
    return [line[2:] if line.startswith('# ')
            else (line[1:] if line.startswith('#') else line)
            for line in lines]


def paragraph_is_fully_commented(lines):
    """Is the paragraph fully commented?"""
    for i, line in enumerate(lines):
        if line.startswith('#'):
            if line.startswith('# %') and is_magic(line):
                return False
            continue
        return i > 0 and _BLANK_LINE.match(line)
    return True


def next_code_is_indented(lines):
    """Is the next unescaped line indented?"""
    for line in lines:
        if _BLANK_LINE.match(line) or _PY_COMMENT.match(line):
            continue
        return _PY_INDENTED.match(line)
    return False


def count_lines_to_next_cell(cell_end_marker, next_cell_start,
                             total, explicit_eoc):
    """How many blank lines between end of cell marker and next cell?"""
    if cell_end_marker < total:
        lines_to_next_cell = next_cell_start - cell_end_marker
        if explicit_eoc:
            lines_to_next_cell -= 1
        if next_cell_start >= total:
            lines_to_next_cell += 1
        return lines_to_next_cell

    return 1


def last_two_lines_blank(source):
    """Are the two last lines blank, and not the third last one?"""
    if len(source) < 3:
        return False
    return (not _BLANK_LINE.match(source[-3]) and
            _BLANK_LINE.match(source[-2]) and
            _BLANK_LINE.match(source[-1]))


class CellReader():
    """A class that can read notebook cells from their text representation"""

    def __init__(self, ext):
        """Create a cell reader with empty content"""
        self.ext = ext
        self.markdown_prefix = ('' if ext in ['.Rmd', '.md'] else
                                "#'" if ext == '.R' else '#')
        self.cell_type = None
        self.language = None
        self.metadata = None
        self.content = []
        self.lines_to_next_cell = 1

    def read(self, lines):
        """Read one cell from the given lines, and return the cell,
        plus the position of the next cell
        """

        # Do we have an explicit code marker on the first line?
        self.metadata_and_language_from_option_line(lines[0])

        # Parse cell till its end and set content, lines_to_next_cell
        pos_next_cell = self.find_cell_content(lines)

        if self.cell_type == 'code':
            new_cell = new_code_cell
        elif self.cell_type == 'markdown':
            new_cell = new_markdown_cell
        else:
            new_cell = new_raw_cell

        if self.lines_to_next_cell != 1:
            self.metadata['lines_to_next_cell'] = self.lines_to_next_cell

        if self.language:
            self.metadata['language'] = self.language

        return new_cell(source='\n'.join(self.content),
                        metadata=self.metadata), pos_next_cell

    def metadata_and_language_from_option_line(self, line):
        """Parse code options on the given line. When a start of a code cell
        is found, self.metadata is set to a dictionary."""
        if self.ext == '.Rmd' and _START_CODE_RMD.match(line):
            self.language, self.metadata = \
                rmd_options_to_metadata(_START_CODE_RMD.findall(line)[0])

        if self.ext == '.md' and _START_CODE_MD.match(line):
            self.language, self.metadata = \
                md_options_to_metadata(_START_CODE_MD.findall(line)[0])

        if self.ext == '.R' and _CODE_OPTION_R.match(line):
            self.language, self.metadata = \
                rmd_options_to_metadata('r ' + _CODE_OPTION_R.findall(line)[0])

        if self.ext in ['.py', '.jl']:
            if _CODE_OPTION_PY.match(line):
                self.language = 'python' if self.ext == '.py' else 'julia'
                self.metadata = json_options_to_metadata(
                    _CODE_OPTION_PY.match(line).group(3))
            if _SIMPLE_START_CODE_PY.match(line):
                self.language = 'python' if self.ext == '.py' else 'julia'
                self.metadata = {}

        if self.metadata and 'language' in self.metadata:
            self.language = self.metadata['language']
            del self.metadata['language']

    def has_explicit_end_marker(self):
        """Should we expect a specific pattern for end of cell?"""

        return self.metadata is not None

    def find_cell_end_rmd(self, lines):
        """Return position of end of cell marker, and position
        of first line after cell"""
        # markdown: (last) two consecutive blank lines
        if self.metadata is None:
            self.cell_type = 'markdown'
            prev_blank = 0
            for i, line in enumerate(lines):
                if (self.ext == '.Rmd' and _START_CODE_RMD.match(line)) or \
                        (self.ext == '.md' and _START_CODE_MD.match(line)):
                    if i > 1 and prev_blank:
                        return i - 1, i, False
                    return i, i, False
                if _BLANK_LINE.match(lines[i]):
                    prev_blank += 1
                elif i > 2 and prev_blank >= 2:
                    return i - 2, i, True
                else:
                    prev_blank = 0
        else:
            self.cell_type = 'code'
            for i, line in enumerate(lines):
                # skip cell header
                if i == 0:
                    continue
                if _END_CODE_MD.match(line):
                    return i, i + 1, True

        # End not found
        return len(lines), len(lines), False

    def find_cell_end_r(self, lines):
        """Return position of end of cell marker, and position
        of first line after cell"""
        if self.metadata is None and lines[0].startswith("#'"):
            self.cell_type = 'markdown'
            for i, line in enumerate(lines):
                if not line.startswith("#'"):
                    if _BLANK_LINE.match(line):
                        return i, i + 1, False
                    return i, i, False

            return len(lines), len(lines), False

        return self.find_cell_end_code(lines, None, _CODE_OPTION_R, "#'")

    def find_cell_end_py(self, lines):
        """Return position of end of cell marker, and position
        of first line after cell"""
        if self.metadata is None and paragraph_is_fully_commented(lines):
            self.cell_type = 'markdown'
            for i, line in enumerate(lines):
                if _BLANK_LINE.match(line):
                    return i, i + 1, False
            return len(lines), len(lines), False

        if self.metadata is not None:
            end_of_cell = self.metadata.get('endofcell', '-')
            cell_end_re = re.compile('^# ' + end_of_cell + r'\s*$')
        else:
            cell_end_re = None

        return self.find_cell_end_code(lines, cell_end_re,
                                       _CODE_OPTION_PY, None)

    def find_cell_end_code(self, lines, cell_end_re,
                           cell_start_re, markdown_prefix):
        """Given that this is a code cell, return position of
        end of cell marker, and position of next cell start"""
        self.cell_type = 'code'
        parser = StringParser('python' if self.ext in ['.py', '.jl'] else 'R')
        for i, line in enumerate(lines):
            # skip cell header
            if self.metadata is not None and i == 0:
                continue

            if parser.is_quoted():
                parser.read_line(line)
                continue

            parser.read_line(line)

            if cell_start_re.match(line) or (
                    markdown_prefix and line.startswith(markdown_prefix)):
                if i > 1 and _BLANK_LINE.match(lines[i - 1]):
                    return i - 1, i, False
                return i, i, False

            if i > 0 and cell_start_re == _CODE_OPTION_PY and \
                    _BLANK_LINE.match(lines[i - 1]) and \
                    _SIMPLE_START_CODE_PY.match(line):
                return i - 1, i, False

            if cell_end_re:
                if cell_end_re.match(line):
                    return i, i + 1, True
            elif _BLANK_LINE.match(line):
                if not next_code_is_indented(lines[i:]):
                    if i > 0:
                        return i, i + 1, False
                    if len(lines) == 1 or _BLANK_LINE.match(lines[1]):
                        return 1, 2, False

        return len(lines), len(lines), False

    def find_cell_end(self, lines):
        """Return position of end of cell marker, and position
        of first line after cell"""
        if self.ext in ['.py', '.jl']:
            return self.find_cell_end_py(lines)
        if self.ext in ['.Rmd', '.md']:
            return self.find_cell_end_rmd(lines)
        return self.find_cell_end_r(lines)

    def find_cell_content(self, lines):
        """Parse cell till its end and set content, lines_to_next_cell.
        Return the position of next cell start"""
        cell_end_marker, next_cell_start, explicit_eoc = \
            self.find_cell_end(lines)

        # Metadata to dict
        if self.metadata is None:
            cell_start = 0
            self.metadata = {}
        else:
            cell_start = 1

        # Cell content
        source = lines[cell_start:cell_end_marker]

        if self.cell_type == 'code':
            if self.ext != '.md' and is_active(self.ext, self.metadata):
                unescape_magic(source, self.language or 'python')
            elif self.ext in ['.py', '.jl', '.R']:
                source = uncomment(source)

        if self.cell_type == 'code' or self.ext in ['.Rmd', '.md']:
            unescape_code_start(source, self.ext, self.language or
                                ('python' if self.ext != '.Rmd' else None))

        if self.cell_type == 'markdown' and self.ext in ['.py', '.jl', '.R']:
            source = self.markdown_unescape(source)

        self.content = source

        # Exactly two empty lines at the end?
        if (self.ext == '.py' and explicit_eoc and
                last_two_lines_blank(source)):
            self.content = source[:-2]
            self.metadata['lines_to_end_of_cell_marker'] = 2

        # Is this a raw cell?
        if not is_active('ipynb', self.metadata) or (
                self.ext == '.md' and self.cell_type == 'code'
                and self.language is None):
            if self.metadata.get('active') == '':
                del self.metadata['active']
            self.cell_type = 'raw'

        # Explicit end of cell marker?
        if (next_cell_start + 1 < len(lines) and
                _BLANK_LINE.match(lines[next_cell_start]) and
                not _BLANK_LINE.match(lines[next_cell_start + 1])):
            next_cell_start += 1
        elif (explicit_eoc and next_cell_start + 2 < len(lines) and
              _BLANK_LINE.match(lines[next_cell_start]) and
              _BLANK_LINE.match(lines[next_cell_start + 1]) and
              not _BLANK_LINE.match(lines[next_cell_start + 2])):
            next_cell_start += 2

        self.lines_to_next_cell = count_lines_to_next_cell(
            cell_end_marker,
            next_cell_start,
            len(lines),
            explicit_eoc)

        return next_cell_start

    def markdown_unescape(self, lines):
        """Remove markdown prefix"""
        prefix = self.markdown_prefix
        prefix_and_space = self.markdown_prefix + ' '
        prefix_and_space_length = len(prefix_and_space)
        prefix_length = len(self.markdown_prefix)

        return [line[prefix_and_space_length:]
                if line.startswith(prefix_and_space)
                else (line[prefix_length:]
                      if line.startswith(prefix) else line)
                for line in lines]



import nbformat
import pytest
import jupytext
from jupytext.compare import compare_notebooks
from .utils import list_r_notebooks

jupytext.file_format_version.FILE_FORMAT_VERSION = {}


@pytest.mark.parametrize('nb_file', list_r_notebooks('.ipynb'))
def test_identity_source_write_read(nb_file):
    """
    Test that writing the notebook with R, and read again,
    is the same as removing outputs
    """

    with open(nb_file) as fp:
        nb1 = nbformat.read(fp, as_version=4)

    R = jupytext.writes(nb1, ext='.R')
    nb2 = jupytext.reads(R, ext='.R')

    compare_notebooks(nb1, nb2)



import os
import sys
import pytest
import jupytext
from jupytext import TextFileContentsManager, readf
from jupytext.compare import compare_notebooks
from .utils import list_all_notebooks, list_py_notebooks

jupytext.file_format_version.FILE_FORMAT_VERSION = {}
jupytext.file_format_version.MIN_FILE_FORMAT_VERSION = {}


@pytest.mark.skipif(isinstance(TextFileContentsManager, str),
                    reason=TextFileContentsManager)
def test_create_contentsmanager():
    TextFileContentsManager()


@pytest.mark.skipif(sys.version_info < (3, 6),
                    reason="unordered dict result in changes in chunk options")
@pytest.mark.skipif(isinstance(TextFileContentsManager, str),
                    reason=TextFileContentsManager)
@pytest.mark.parametrize('nb_file', list_all_notebooks('.ipynb'))
def test_load_save_rename(nb_file, tmpdir):
    tmp_ipynb = 'notebook.ipynb'
    tmp_rmd = 'notebook.Rmd'

    cm = TextFileContentsManager()
    cm.default_jupytext_formats = 'ipynb,Rmd'
    cm.root_dir = str(tmpdir)

    # open ipynb, save Rmd, reopen
    nb = readf(nb_file)
    cm.save(model=dict(type='notebook', content=nb), path=tmp_rmd)
    nb_rmd = cm.get(tmp_rmd)
    compare_notebooks(nb, nb_rmd['content'])

    # save ipynb
    cm.save(model=dict(type='notebook', content=nb), path=tmp_ipynb)

    # rename ipynb
    cm.rename(tmp_ipynb, 'new.ipynb')
    assert not os.path.isfile(str(tmpdir.join(tmp_ipynb)))
    assert not os.path.isfile(str(tmpdir.join(tmp_rmd)))

    assert os.path.isfile(str(tmpdir.join('new.ipynb')))
    assert os.path.isfile(str(tmpdir.join('new.Rmd')))


@pytest.mark.skipif(sys.version_info < (3, 6),
                    reason="unordered dict result in changes in chunk options")
@pytest.mark.skipif(isinstance(TextFileContentsManager, str),
                    reason=TextFileContentsManager)
@pytest.mark.parametrize('nb_file', list_py_notebooks('.ipynb'))
def test_load_save_rename_nbpy(nb_file, tmpdir):
    tmp_ipynb = 'notebook.ipynb'
    tmp_nbpy = 'notebook.nb.py'

    cm = TextFileContentsManager()
    cm.default_jupytext_formats = 'ipynb,nb.py'
    cm.root_dir = str(tmpdir)

    # open ipynb, save nb.py, reopen
    nb = readf(nb_file)
    cm.save(model=dict(type='notebook', content=nb), path=tmp_nbpy)
    nbpy = cm.get(tmp_nbpy)
    compare_notebooks(nb, nbpy['content'])

    # save ipynb
    cm.save(model=dict(type='notebook', content=nb), path=tmp_ipynb)

    # rename nbpy
    cm.rename(tmp_nbpy, 'new.nb.py')
    assert not os.path.isfile(str(tmpdir.join(tmp_ipynb)))
    assert not os.path.isfile(str(tmpdir.join(tmp_nbpy)))

    assert os.path.isfile(str(tmpdir.join('new.ipynb')))
    assert os.path.isfile(str(tmpdir.join('new.nb.py')))


@pytest.mark.skipif(sys.version_info < (3, 6),
                    reason="unordered dict result in changes in chunk options")
@pytest.mark.skipif(isinstance(TextFileContentsManager, str),
                    reason=TextFileContentsManager)
@pytest.mark.parametrize('nb_file', list_py_notebooks('.ipynb'))
def test_load_save_rename_nbpy_default_config(nb_file, tmpdir):
    tmp_ipynb = 'notebook.ipynb'
    tmp_nbpy = 'notebook.nb.py'

    cm = TextFileContentsManager()
    cm.default_jupytext_formats = 'ipynb'
    cm.root_dir = str(tmpdir)

    # open ipynb, save nb.py, reopen
    nb = readf(nb_file)
    cm.save(model=dict(type='notebook', content=nb), path=tmp_nbpy)
    nbpy = cm.get(tmp_nbpy)
    compare_notebooks(nb, nbpy['content'])

    # open ipynb
    nbipynb = cm.get(tmp_ipynb)
    compare_notebooks(nb, nbipynb['content'])

    # save ipynb
    cm.save(model=dict(type='notebook', content=nb), path=tmp_ipynb)

    # rename nbpy
    cm.rename(tmp_nbpy, 'new.nb.py')
    assert not os.path.isfile(str(tmpdir.join(tmp_ipynb)))
    assert not os.path.isfile(str(tmpdir.join(tmp_nbpy)))

    assert os.path.isfile(str(tmpdir.join('new.ipynb')))
    assert os.path.isfile(str(tmpdir.join('new.nb.py')))

    # rename ipynb
    cm.rename('new.ipynb', tmp_ipynb)
    assert os.path.isfile(str(tmpdir.join(tmp_ipynb)))
    assert not os.path.isfile(str(tmpdir.join(tmp_nbpy)))

    assert not os.path.isfile(str(tmpdir.join('new.ipynb')))
    assert os.path.isfile(str(tmpdir.join('new.nb.py')))
